import multiprocessing
import warnings
from typing import Iterable, Optional, Union

from ...exec_model import ModelObject
from ._ga import GeneticAlgorithmContinue, GeneticAlgorithmInit


def _check_optional_arguments(
    x_id: Union[int, Iterable[int]],
    options: Optional[dict],
) -> None:
    warnings.warn(
        "This function will no longer be supported. Use biomass.optimize() for parameter estimation.",
        FutureWarning,
    )

    if options is None:
        pass
    elif isinstance(options, dict):
        if options["local_search_method"].lower() not in ["mutation", "powell", "de"]:
            raise ValueError(
                f"'{options['local_search_method']}': "
                "Invalid local_search_method. Should be one of ['mutation', 'Powell', 'DE']"
            )
        elif (
            not isinstance(x_id, int)
            and options["local_search_method"].lower() == "de"
            and options["workers"] != 1
        ):
            raise AssertionError(
                "daemonic processes are not allowed to have children. Set options['workers'] to 1."
            )
    else:
        raise TypeError("options must be dict or None.")


def optimize(
    model: ModelObject,
    x_id: Union[int, Iterable[int]],
    options: Optional[dict] = None,
) -> None:
    """
    Estimate model parameters from experimental data.
    Parameters
    ----------
    model : ModelObject
        Model for parameter estimation.
    x_id : int or Iterable[int]
        Index (indices) of parameter set to estimate.
    options : dict, optional
        * popsize : int (default: 5)
            A multiplier for setting the total population size.
            The population has ``popsize`` * len(`search_param`) individuals.
        * max_generation : int (default: 10000)
            Stop optimization if Generation > ``max_generation``.
        * initial_threshold : float (default: 1e12)
            Threshold on objective function value used to generate initial population.
            Default value is 1e12 (numerically solvable).
        * allowable_error : float (default: 0.0)
            Stop optimization if Best Fitness <= ``allowable_error``.
        * local_search_method : str (default: 'mutation')
            Method used in local search. Should be one of:
            * 'mutation' : NDM/MGG
            * 'Powell' : Modified Powell method
            * 'DE' : Differential Evolution (strategy: `best2bin`)
        * n_children : int (default: 1000)
            (``local_search_method`` == 'mutation') The number of children generated in NDM/MGG.
        * maxiter : int (default: 10)
            (``local_search_method`` in ['Powell', 'DE']) The maximum number of iterations
            over which the entire population is evolved.
        * workers : int (default: -1 if isinstance(x_id, `int`) else 1)
            (``local_search_method`` == 'DE') The population is subdivided into workers sections and
            evaluated in parallel (uses multiprocessing.Pool). Supply -1 to use
            all available CPU cores. Set workers to 1 when searching multiple
            parameter sets simultaneously.
        * overwrite : bool (default: :obj:`False`)
            If :obj:`True`, the out/n folder will be overwritten.
    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import Model, optimize
    >>> model = Model(Nakakuki_Cell_2010.__package__).create()
    >>> optimize(
    ...     model, x_id=range(1, 11),
    ...     options={
    ...         'max_generation': 10000,
    ...         'allowable_error': 0.5
    ...     }
    ... )
    """
    if options is None:
        options = {}
    options.setdefault("popsize", 5)
    options.setdefault("max_generation", 10000)
    options.setdefault("initial_threshold", 1e12)
    options.setdefault("allowable_error", 0.0)
    options.setdefault("local_search_method", "mutation")
    options.setdefault("n_children", 1000)
    options.setdefault("maxiter", 10)
    options.setdefault("workers", -1 if isinstance(x_id, int) else 1)
    options.setdefault("overwrite", False)

    _check_optional_arguments(x_id, options)

    ga_init = GeneticAlgorithmInit(model, **options)
    if isinstance(x_id, int):
        ga_init.run(x_id)
    elif all([isinstance(i, int) for i in x_id]):
        n_proc = max(1, multiprocessing.cpu_count() - 1)
        with multiprocessing.Pool(processes=n_proc) as p:
            for _ in p.imap_unordered(ga_init.run, x_id):
                pass
    else:
        raise TypeError("`x_id` type must be either int or Iterable[int].")


def optimize_continue(
    model: ModelObject,
    x_id: Union[int, Iterable[int]],
    options: Optional[dict] = None,
) -> None:
    """
    Continue running optimization from where you stopped in the last parameter search.
    Parameters
    ----------
    model : ModelObject
        Model for parameter estimation.
    x_id : int or Iterable[int]
        Index (indices) of parameter set to estimate.
    options : dict, optional
        * popsize : int (default: 5)
            A multiplier for setting the total population size.
            The population has ``popsize`` * len(`search_param`) individuals.
        * max_generation : int (default: 15000)
            Stop optimization if Generation > ``max_generation``.
        * initial_threshold : float (default: 1e12)
            Threshold on objective function value used to generate initial population.
            Default value is 1e12 (numerically solvable).
        * allowable_error : float (default: 0.0)
            Stop optimization if Best Fitness <= ``allowable_error``.
        * local_search_method : str (default: 'mutation')
            Method used in local search. Should be one of:
            * 'mutation' : NDM/MGG
            * 'Powell' : Modified Powell method
            * 'DE' : Differential Evolution (strategy: `best2bin`)
        * n_children : int (default: 1000)
            (``local_search_method`` == 'mutation') The number of children generated in NDM/MGG.
        * maxiter : int (default: 10)
            (``local_search_method`` in ['Powell', 'DE']) The maximum number of iterations
            over which the entire population is evolved.
        * workers : int (default: -1 if isinstance(x_id, `int`) else 1)
            (``local_search_method`` == 'DE') The population is subdivided into workers sections
            and evaluated in parallel (uses multiprocessing.Pool). Supply -1 to use
            all available CPU cores. Set workers to 1 when searching multiple
            parameter sets simultaneously.
        * p0_bounds : list of floats (default: [0.1, 10.0])
            Generate initial population using best parameter values in the last parameter search.
            * `lower_bound` = po_bounds[0] * `best_parameter_value`
            * `upper_bound` = p0_bounds[1] * `best_parameter_value`
    Examples
    --------
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> from biomass import Model, optimize_continue
    >>> model = Model(Nakakuki_Cell_2010.__package__).create()
    >>> optimize_continue(
    ...     model, x_id=range(1, 11),
    ...     options={
    ...         'max_generation': 20000,
    ...         'allowable_error': 0.5
    ...     }
    ... )
    """
    if options is None:
        options = {}
    options.setdefault("popsize", 5)
    options.setdefault("max_generation", 15000)
    options.setdefault("initial_threshold", 1e12)
    options.setdefault("allowable_error", 0.0)
    options.setdefault("local_search_method", "mutation")
    options.setdefault("n_children", 1000)
    options.setdefault("maxiter", 10)
    options.setdefault("workers", -1 if isinstance(x_id, int) else 1)
    options.setdefault("p0_bounds", [0.1, 10.0])

    _check_optional_arguments(x_id, options)

    ga_continue = GeneticAlgorithmContinue(model, **options)
    if isinstance(x_id, int):
        ga_continue.run(x_id)
    elif all([isinstance(i, int) for i in x_id]):
        n_proc = max(1, multiprocessing.cpu_count() - 1)
        with multiprocessing.Pool(processes=n_proc) as p:
            for _ in p.imap_unordered(ga_continue.run, x_id):
                pass
    else:
        raise TypeError("`x_id` type must be either int or Iterable[int].")



import os
import shutil
import sys
import warnings
from typing import Callable, List, Union

import numpy as np

from ..exec_model import ModelObject

DIRNAME = "_tmp"


class _Logger(object):
    """
    Duplicate stdout to _tmp/optimization.log.
    """

    def __init__(self, model_path: str, x_id: int, disp_here: bool):
        self.disp_here = disp_here
        self.terminal = sys.stdout
        self.log = open(
            os.path.join(model_path, "out", DIRNAME + str(x_id), "optimization.log"),
            mode="w",
            encoding="utf-8",
        )

    def write(self, message: str):
        if self.disp_here:
            self.terminal.write(message)
        self.log.write(message)


class Optimizer(object):
    """
    Use an external optimization method for parameterization of a mechanistic model.

    Attributes
    ----------
    model : ModelObject
        The BioMASS model object.
    optimize : Callable
        The optimizer, e.g., :func:`scipy.optimize.differential_evolution`.
    x_id : int
        Index of parameter set to estimate.
    disp_here: bool (default: False)
        Whether to show the evaluated *objective* at every iteration.

    Examples
    --------
    >>> from scipy.optimize import differential_evolution
    >>> from biomass import Model
    >>> from biomass.estimation import ExternalOptimizer
    >>> from biomass.models import Nakakuki_Cell_2010
    >>> model = Model(Nakakuki_Cell_2010.__package__).create()
    >>> param_idx = 1
    >>> optimizer = ExternalOptimizer(model, differential_evolution, param_idx)
    >>> def obj_fun(x):
    ...    '''Objective function to be minimized.'''
    ...    return optimizer.get_obj_val(x)
    >>> res = optimizer.minimize(
    ...     obj_fun,
    ...     [(0, 1) for _ in range(len(model.problem.bounds))],
    ...     strategy="best1bin",
    ...     maxiter=50,
    ...     tol=1e-4,
    ...     mutation=0.1,
    ...     recombination=0.5,
    ...     disp=True,
    ...     polish=False,
    ...     workers=-1,
    ... )

    differential_evolution step 1: f(x)= 5.19392\n
    differential_evolution step 2: f(x)= 2.32477\n
    differential_evolution step 3: f(x)= 1.93583\n
    ...\n
    differential_evolution step 50: f(x)= 0.519774\n

    >>> from biomass import run_simulation
    >>> param_values = model.problem.gene2val(res.x)
    >>> optimizer.import_solution(param_values)
    >>> run_simulation(model, viz_type=str(param_idx))
    """

    def __init__(
        self,
        model: ModelObject,
        optimize: Callable,
        x_id: int,
        disp_here: bool = False,
    ):
        self.model = model
        self.optimize = optimize
        self.x_id = x_id
        self.disp_here = disp_here

        self.savedir = os.path.join(self.model.path, "out", f"{self.x_id}")
        if os.path.isdir(self.savedir):
            raise ValueError(
                f"out{os.sep}{self.x_id} already exists in {self.model.path}. "
                "Use another parameter id."
            )
        else:
            os.makedirs(self.savedir)
        os.makedirs(os.path.join(self.model.path, "out", DIRNAME + str(self.x_id)), exist_ok=True)
        self.default_stdout = sys.stdout

    def minimize(self, *args, **kwargs):
        """
        Execute the external optimizer.
        """
        os.makedirs(os.path.join(self.model.path, "out", DIRNAME + str(self.x_id)), exist_ok=True)
        try:
            sys.stdout = _Logger(self.model.path, self.x_id, self.disp_here)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                res = self.optimize(*args, **kwargs)
            return res
        finally:
            sys.stdout = self.default_stdout

    def _get_n_iter(self) -> int:
        n_iter: int = 0
        path_to_log = os.path.join(self.savedir, "optimization.log")
        with open(path_to_log, mode="r", encoding="utf-8") as f:
            log_file = f.readlines()
        for message in log_file:
            if len(message.strip()) > 0:
                n_iter += 1
        return n_iter

    def import_solution(self, x: Union[np.ndarray, List[float]], cleanup: bool = True) -> None:
        """
        Import the solution of the optimization to the model.
        The solution vector `x` will be saved to `path_to_model`/out/`x_id`/.
        Use ``biomass.run_simulation`` to visualize the optimization result.

        Parameters
        ----------
        x : Union[np.ndarray, List[float]]
            The solution of the optimization.
        cleanup : bool (default: True)
            If True (default), delete the temporary folder after the optimization is finished.
        """
        shutil.move(
            os.path.join(self.model.path, "out", DIRNAME + str(self.x_id), "optimization.log"),
            self.savedir,
        )

        best_fitness: float = self.model.problem.objective(x)
        n_iter = self._get_n_iter()
        np.save(os.path.join(self.savedir, "best_fitness"), best_fitness)
        np.save(os.path.join(self.savedir, "count_num"), n_iter)
        np.save(os.path.join(self.savedir, "generation"), n_iter)
        np.save(os.path.join(self.savedir, f"fit_param{n_iter}"), x)
        if cleanup:
            shutil.rmtree(os.path.join(self.model.path, "out", DIRNAME + str(self.x_id)))



"""BioMASS - A Python Framework for Modeling and Analysis of Signaling Systems"""

from .core import *
from .result import OptimizationResults
from .version import __version__

__author__ = __maintainer__ = "Hiroaki Imoto"
__email__ = "himoto@protein.osaka-u.ac.jp"
